


/*
 *  $Id: dmch.h,v 1.22 1994/07/26 18:08:51 gropp Exp gropp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/* 
   These are the Chameleon-specific macros.

   See the readme file for details
 */


#ifndef _DMCH_INCLUDED
#define _DMCH_INCLUDED

#define MPIDPATCHLEVEL 1.1
#define MPIDTRANSPORT "Intel NX"
/* 
   Redefine the names of the async communication types (typedefed in 
   comm$(COMM).h) so that a Chameleon program can include commmpi.h which
   includes this file (through the mpi.h file)  

   I'm not sure that this does what I need, but it doesn't hurt.
 */
#define ASYNCRecvId_t MPIDCH_ASYNCRecvId_t
#define ASYNCSendId_t MPIDCH_ASYNCSendId_t

/* 
   When we compile the device, we want to include all of the device code,
   but when we compile user code, we don't want to require that they load 
   the defintions in either tools.h or comm/comm.h.  
 */

#ifdef MPID_DEVICE_CODE
#include <stdio.h>
extern int __NUMNODES, __MYPROCID;
#else
/* 
   These are used to provide a simple type for the rest of the MPI (API)
   code that needs to have this information when handling the device 
   structures
 */
#endif

#include "mpi.h"
#include "dmpiatom.h"

/* 
   This packet size should be selected such that
   (s + r*(n+h)) + c*n \approx (s+r*n) + s+r*h
   where s = latency, r = time to send a byte, n = total message length, 
   h = header size, and c = time to copy a byte.  This condition reduces to
   c n \approx s
   For a typical system with
   s = 30us
   c = .03us/byte
   this gives
   n = s / c = 30 us / (.03us/byte) = 1000 bytes

   When the message does not fit into a single packet, ALL of the message
   should be placed in the "extension" packet (see below).  This removes 
   an extra copy from the code.
 */
#define MPID_PACKET_SIZE 1024


/*
  The implementation reserves some message tags.

  (An optimization is to allow the use of all but a few very large tags
  for messages in the initial communicator, thus eliminating a separate
  header.  Messages in a different communicator would be sent on a reserved
  set of tags.  An alternate is to use the Chameleon tags for communicator
  types, making the message-passing system handle the queueing of messages
  by communicator.  Note that if tags are used to separate communicators,
  message-passing systems that have stream semantics, like EUI-H, will
  fail to operate correctly.  Another approach is to make the tags a 
  combination of tag, context, and source.)

  PT2PT_TAG is the tag for short messages and the headers of long messages
  PT2PT2_TAG(source) is the tag for longer messages (by source).  This permits
  the header messages to be freely received into preallocated buffers, and
  for long messages to be received directly into user-buffers.

  The mode field is overloaded for the synchronous case because we support
  NONBLOCKING, SYNCHRONOUS sends; thus, there can be a variety of outstanding
  synchronous sends at any time, and we have to match them up.

  We do this by making the mode field look like this:

  <syncreqnum><modetype>

  The mode field is required because, while there are different sends for each
  mode, there is only one kind of receive, and hence we need the mode field
  to decide what to do.  In fact, our only need is to handle MPI_SYNCHRONOUS
  sends.
 */
#define MPID_PT2PT_TAG 0
#define MPID_PT2PT2_TAG(src) (1+(src))

/* Whether an operation should block or not */
typedef enum { MPID_NOTBLOCKING = 0, MPID_BLOCKING } MPID_BLOCKING_TYPE;

/*
   The mode field contains an ID if the mode is SYNCHRONOUS 

   This is a very simple, open, single packet structure.  Additional
   performance, particularly for short messages, can be obtained by:
       mode is two bits; if sync or sync_ack, next nn bits are id
       lrank is 8 or 14 bits (depending on size of system; the reason
       for 14 is made clear below).
       len is provided only if a "large message" bit is set; otherwise,
       the length is computed from what is actually delivered.
       context_id is 16 bits
   With these changes, the short, non-synchronous send packet 
   is only 31 (tag) + 1 (short) + 2 (mode) + 16 context_id + 14 (lrank) =
   64 bits.  The packet defined here is 160 bits.

   Similar games can be played for other design points.  Note that the 
   lrank could be determined by looking up the absolute rank in the 
   matching context_id; this approach may be cost effective if many small
   messages are sent on a slow system.
 */

/* 
   Here are definitions from mpi_bc.h ...
typedef enum { 
    MPIR_MODE_STANDARD = 0, 
    MPIR_MODE_READY = 1, 
    MPIR_MODE_SYNCHRONOUS = 2, 
    MPIR_MODE_RECV = 3,
    MPIR_MODE_SYNC_ACK = 4
 */

#define MPID_MODE_MASK  0xf
#define MPID_MODE_BITS    4

/* Mode bits are:
   0x1 - Ready
   0x2 - Synchronous
   0x4 - XDR (two-ended transmission protocol)
   0x8 - Long message (not yet used)

   THESE MUST BE THE SAME AS THE API USES! (see include/mpi_bc.h)
   Note that the API just uses the first two modes; XDR is indicated
   in a separate word.  Also node that the the API stores the mode and the
   sync_id in a single word; see MPIR_MODE_xxxx below.
 */
#define MPID_MODE_READY 0x1
#define MPID_MODE_SYNC  0x2
#define MPID_MODE_XDR   0x4
#define MPID_MODE_LONG  0x8
#define MPID_MODE_SYNC_ACK (MPID_MODE_SYNC | MPID_MODE_READY)
#define MPIR_MODE_IS_SYNC(mpid) ((mpid)->mode & MPIR_MODE_SYNCHRONOUS)
#define MPIR_MODE_SYNC_ID(mpid) ((mpid)->mode >> MPID_MODE_BITS)

/* 
   The compressed packet format trades a slightly smaller packet for 
   more complicated code to access the entries.  In addition, it 
   does not send the id of the synchronous message unless the message type
   is indeed synchronous.
 */
#ifdef USE_COMPRESSED_PACKET
/* 
   Long messages have no local data; we let them include the sync_id 
   as the last field (which need not be sent!)
 */
typedef struct {
    int len;                    /* TOTAL length of message in BYTES */
    int sync_id;                /* Id of a synchronous message */
    } MPID_PACKET_LONG;
typedef struct {
    int sync_id;                /* Id of a synchronous message */
	char buffer[MPID_PACKET_SIZE];
				/* Maximum message size for short messages */
    } MPID_PACKET_SYNC;
typedef struct {
    unsigned mode:4;            /* Mode */
    unsigned context_id:16;     /* Context_id */
    unsigned lrank:12;          /* Local rank in sending context */
    int tag;                    /* Tag is a full 32 bits */
    union {
	char buffer[MPID_PACKET_SIZE];
				/* Maximum message size for short messages */
	MPID_PACKET_SYNC s_short;  /* Short, synchronous message */
	MPID_PACKET_LONG s_long;   /* Long message, may be synchronous */
	};
    } MPID_PACKET;

/* Header_Len is just the length of the envelope of MPID_PACKET */
#define MPID_HEADER_LEN (sizeof(MPID_PACKET)-MPID_PACKET_SYNC)

#define MPID_HEADER_INTS (MPID_HEADER_LEN/sizeof(int))  
                           /* Number of ints in the header */
#define MPID_SYNC_GET_ID(pkt)    ((pkt)->sync_id)
#define MPID_SYNC_SET_ID(pkt,id) (pkt)->sync_id = id
#define MPID_MODE_GET_SYNC_WITH_ID(pkt) \
    ((pkt)->sync_id << MPID_MODE_BITS | (pkt)->mode )
#define MPID_MODE_IS_SYNC(pkt)   ((pkt)->mode & MPID_MODE_SYNC)
#define MPID_MODE_IS_READY(pkt)  ((pkt)->mode & MPID_MODE_READY)
#define MPID_MODE_IS_SYNCACK(pkt) \
    ((pkt)->mode == MPID_MODE_SYNC_ACK)
#define MPID_MODE_HAS_XDR(pkt)   ((pkt)->mode & MPID_MODE_XDR)

#define MPID_MSG_LEN(pkt) \
    ((pkt)->mode & MPID_MODE_LONG ? (ptk)->len : infocount() - MPID_HEADER_LEN)
#define MPID_MSG_SET_LEN(pkt,s) \
    ((pkt)->mode & MPID_MODE_LONG ? (ptk)->len = s : 0)
#else
typedef struct {
    int len,			/* TOTAL length of message in BYTES */
        tag,			/* Message tag */
        context_id,		/* Internal communicator ID */
        mode,                   /* mode (standard, ready, synchronous,
				   sync_ack) */
        lrank;                  /* rank in sending context */
    char buffer[MPID_PACKET_SIZE];
				/* Maximum message size for short messages */
    } MPID_PACKET;

/* HeaderLen is just the length of the envelope of MPID_PACKET */
#define MPID_HEADER_LEN (sizeof(MPID_PACKET)-MPID_PACKET_SIZE)

/* Number of ints in the header */
#define MPID_HEADER_INTS 5    
#define MPID_SYNC_GET_ID(pkt)    ((pkt)->mode >> MPID_MODE_BITS)
#define MPID_SYNC_SET_ID(pkt,id) ((pkt)->mode |= ((id) << MPID_MODE_BITS))
#define MPID_MODE_GET_SYNC_WITH_ID(pkt) (pkt)->mode
#define MPID_MODE_IS_SYNC(pkt)   \
         (((pkt)->mode & MPID_MODE_MASK) == MPIR_MODE_SYNCHRONOUS)
#define MPID_MODE_IS_READY(pkt)  ((pkt)->mode & MPID_MODE_READY)
#define MPID_MODE_IS_SYNCACK(pkt) \
    ((pkt)->mode & MPID_MODE_MASK == MPID_MODE_SYNC_ACK)
#define MPID_MODE_HAS_XDR(pkt)   ((pkt)->mode & MPID_MODE_XDR)

#define MPID_MSG_LEN(pkt) (pkt)->len
#define MPID_MSG_SET_LEN(pkt,s) (pkt)->len = s
#endif

/*
   Another option would be for the device handle to contain the initial
   packet.  I have NOT do this so as to keep down the size of the device
   handle (since I want relatively large packets)

   This needs to be structured so that there is just enough here for mpir.h.
 */
typedef struct {
    MPIR_BOOL     done;             /* done is set when the message has
				       been sent and, if a SYNC mode,
				       the ack has been received */
    int           is_non_blocking;
    int sid;              /* Id of non-blocking send, if used.
				       0 if no non-blocking send used, 
				       or if non-blocking send has completed */
        /* The following describes the buffer to be sent */
    void          *start;
    int           bytes_as_contig;
    MPIR_NODETYPE dataelement;      /* Will eventually hold datatype for
				       heterogeneous systems */
    } MPID_SHANDLE;

typedef struct {
    MPIR_BOOL     done;
    int           is_non_blocking;
    int rid;              /* Id of non-blocking recv, if used.
				       0 if no non-blocking recv used */
    char          *temp;            /* Holds body of unexpected message */
    int           mode;             /* mode bits and sequence number; needed
				       for unexpected messages */
    int           from;             /* Absolute process number that sent
				       message; used only for SYNC ack */

        /* The following describes the buffer to be received */
    void          *start;
    int           bytes_as_contig;
    MPIR_NODETYPE dataelement;      /* Will eventually hold datatype for
				       heterogeneous systems */

    } MPID_RHANDLE;

#define MPID_MIN(a,b) ((a) < (b) ? (a) : (b))

#define MPID_CAN_SEND_CONTIG

/* 
   Since allocation is done by placing the device structure directly into
   the MPIR_?HANDLE, we don't need to allocate space.  We do, however, take
   this opportunity to initialize it...

   The ...reuse... versions are for persistant handles (e.g., MPI_Send_init)
 */
#define MPID_alloc_send_handle( a ) {(a)->done = MPIR_NO;}
#define MPID_alloc_recv_handle( a ) {(a)->done = MPIR_NO;(a)->temp  = 0;}
#define MPID_free_send_handle( a )  
#define MPID_free_recv_handle( a )  if ((a)->temp  ) {FREE((a)->temp);}
#define MPID_reuse_send_handle( a ) {(a)->done = MPIR_NO;}
#define MPID_reuse_recv_handle( a ) {(a)->done = MPIR_NO;(a)->temp  = 0;}
#define MPID_set_send_is_nonblocking( a, v ) (a)->is_non_blocking = v
#define MPID_set_recv_is_nonblocking( a, v ) (a)->is_non_blocking = v

/* Contact with the device layer is made here.  These call the
   routines to actually process a message (see dmch.c) 

   We use different names to enable the use of a multi-protocol system
   (planned for future support)
 */
#define MPID_post_send(dmpi_send_handle) \
    MPID_NX_post_send(dmpi_send_handle) 
#define MPID_post_recv(dmpi_recv_handle, is_available ) \
    MPID_NX_post_recv(dmpi_recv_handle, is_available ) 
#define MPID_blocking_recv(dmpi_recv_handle ) \
    MPID_NX_blocking_recv(dmpi_recv_handle) 
#define MPID_complete_recv(dmpi_recv_handle) \
    MPID_NX_complete_recv(dmpi_recv_handle) 
#define MPID_complete_send(dmpi_send_handle,status) \
    MPID_NX_complete_send(dmpi_send_handle,status) 
#define MPID_check_device( blocking ) \
    MPID_NX_check_device( blocking )
#define MPID_Iprobe( tag, source, context_id, flag, status ) \
    MPID_NX_Iprobe( tag, source, context_id, flag, status ) 
#define MPID_Probe( tag, source, context_id, status ) \
    MPID_NX_Probe( tag, source, context_id, status ) 

#define MPID_NODE_NAME( name, len ) \
    MPID_NX_Node_name( name, len )
#define MPID_Version_name( name ) MPID_NX_Version_name( name )
#define MPID_WTIME()         MPID_NX_Wtime()
#define MPID_WTICK()         MPID_NX_Wtick()
#define MPID_INIT(argc,argv) MPID_NX_Init( argc, argv )
#define MPID_END()           MPID_NX_End()

#define MPID_ABORT( errorcode ) MPID_NX_Abort( errorcode );

#define MPID_CANCEL( r ) MPID_NX_Cancel( r )

#define MPID_Myrank( rank ) MPID_NX_Myrank( rank )
#define MPID_Mysize( size ) MPID_NX_Mysize( size )

/* thread locking.  Single-thread devices will make these empty 
   declarations */
#define MPID_THREAD_LOCK(comm)
#define MPID_THREAD_UNLOCK(comm)
#define MPID_THREAD_LOCK_INIT(comm)
#define MPID_THREAD_LOCK_FINISH(comm)

/* This device prefers that the data be prepacked (at least for now) */
#define MPID_PACK_IN_ADVANCE
#define MPID_RETURN_PACKED

/* Forward refs */
#ifdef __STDC__
extern void MPID_NX_Abort( );
extern void MPID_NX_Myrank( int * ), MPID_NX_Mysize( int * ), 
            MPID_NX_End(void);
/* 
extern int MPID_NX_post_send( MPIR_SHANDLE * ), 
           MPID_NX_blocking_recv( MPIR_SHANDLE *), 
           MPID_NX_complete_recv( MPIR_SHANDLE *),
           MPID_NX_complete_send( MPIR_SHANDLE *, MPI_Status *);
 */
/* MPID_NX_check_device( blocking ), 
   MPID_NX_Iprobe( tag, source, context_id, flag, status ), 
   MPID_NX_Probe( tag, source, context_id, status ) 
 */
#else
extern void MPID_NX_Abort();
extern void MPID_NX_Myrank(), MPID_NX_Mysize(), MPID_NX_End();
#endif

#endif

#ifdef MPID_DEVICE_CODE
/* Device-only information */
#include "../../include/dmpiatom.h"

/* End of code included only when building the ADI routines */
#endif

/* For heterogeneous support --- NOT YET FULLY IMPLEMENTED 
   (fully means that there is some code that is not yet used).
   This provides information on how data should be communicated to 
   a processor.  The approach is to only convert data when
   the formats are different on the source and destination, and then to
   use byte-swapping code rather than xdr on the SENDER where possible.
   In a heterogeneous environment, the receiver need only check for 
   a sender that had to use MPID_H_XDR; otherwise, the received data is
   already in the correct format.

   None of this code is executed or included in a homogeneous environment
 */
#ifndef MPID_H_INC
#define MPID_H_INC
typedef enum { MPID_H_NONE = 0, 
		   MPID_H_LSB, MPID_H_MSB, MPID_H_XDR } MPID_H_TYPE;
/* 
   The MPID_INFO structure is acquired from each node and used to determine
   the format for data that is sent 
 */
typedef struct {
    MPID_H_TYPE byte_order;
    } MPID_INFO;
extern MPID_INFO *MPID_procinfo;
extern MPID_H_TYPE MPID_byte_order;

extern void (*MPID_ErrorHandler)();
extern void MPID_DefaultErrorHandler();

#if defined(p4) || defined(pvm) || defined(pvm3) || defined(MPID_HAS_HETERO)
#ifndef MPID_HAS_HETERO
#define MPID_HAS_HETERO
#endif
extern int MPID_IS_HETERO;
#define MPID_Dest_byte_order(dest) MPID_NX_Dest_byte_order(dest)
#endif

#endif

