/*
 *  $Id: waitany.c,v 1.10 1994/08/12 20:39:22 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      See COPYRIGHT in top-level directory.
 */


#ifndef lint
static char vcid[] = "$Id: waitany.c,v 1.10 1994/08/12 20:39:22 gropp Exp $";
#endif /* lint */
#include "mpiimpl.h"
#include "mpisys.h"

/*@
    MPI_Waitany - Waits for any specified send or receive to complete

Input Parameters:
. count - list length (integer) 
. array_of_requests - array of requests (array of handles) 

Output Parameters:
. index - index of handle for operation that completed (integer).  In the
range 0 to count-1.  In Fortran, the range is 1 to count.
. status - status object (Status) 
@*/
int MPI_Waitany(count, array_of_requests, index, status )
int         count;
MPI_Request array_of_requests[];
int         *index;
MPI_Status  *status;
{
    int i, errno = MPI_SUCCESS;
    MPI_Request request;
    int numleft;

    *index = MPI_UNDEFINED;

    /* Check for the trivial case of nothing to do. */
    for (i=0; i<count; i++) {
	if (array_of_requests[i] && array_of_requests[i]->chandle.active) 
	    break;
	}
    if (i == count) return errno;

    while (1) {
	for (i = 0; i < count; i++)
	    {
	    /* Skip over null handle.  We need this for handles generated
	       when MPI_PROC_NULL is the source or destination of an 
	       operation */
	    request = array_of_requests[i];
	    if (!request || !request->chandle.active ) continue;

	    if (request->chandle.completed) {
		*index = i;
		if ( request->type == MPIR_RECV ) {
		    MPID_complete_recv( &request->rhandle );
		    
		    status->MPI_SOURCE	   = request->rhandle.source;
		    status->MPI_TAG	   = request->rhandle.tag;
		    status->count	   = request->rhandle.actcount;  
		    status->count_in_bytes = request->rhandle.totallen;
#ifdef MPID_RETURN_PACKED
		    if (request->rhandle.bufpos) 
			errno = MPIR_UnPackMessage( request->rhandle.bufadd, 
						   request->rhandle.count, 
						   request->rhandle.datatype, 
						   request->rhandle.source,
						   request );
#endif
		    }
		else {
		  MPID_complete_send( &request->shandle, 
				     status );
#if defined(MPID_PACK_IN_ADVANCE) || defined(MPID_HAS_HETERO)
		  if (request->shandle.bufpos && 
		      (errno = MPIR_SendBufferFree( request ))){
		    MPIR_ERROR( MPI_COMM_NULL, errno, 
			       "Could not free allocated send buffer in MPI_WAITANY" );
		  }
#endif

		}

	        /* Spec requires, but... */
		if (!request->chandle.persistent) {
		    if (--request->chandle.datatype->ref_count <= 0) {
			MPIR_Type_free( &request->chandle.datatype );
			}
		    MPI_Request_free( &array_of_requests[i] ); 
		    array_of_requests[i]    = NULL;
		    }
		else {
		    request->chandle.active    = 0;
		    request->chandle.completed = MPIR_NO;
		    if (request->type == MPIR_RECV) {
			MPID_reuse_recv_handle(&request->rhandle.dev_rhandle );
			}
		    else {
			MPID_reuse_send_handle(&request->shandle.dev_shandle );
			}
		    }
		return MPI_SUCCESS;
		}
	    }
	/* If nothing was ready, do a blocking wait on the device for 
	   anything to arrive; we'll then re-check the requests */
	MPID_check_device( MPID_BLOCKING );
	}
}

